
#ifndef __WAVE_DRV_H__
#define __WAVE_DRV_H__

//===========================================
// wave format definition
//===========================================
typedef enum 
{
     PCM_TYPE_M8,
     PCM_TYPE_M16,
     PCM_TYPE_S8,
     PCM_TYPE_S16
} PCM_TYPE, *PPCM_TYPE;



//===========================================
// wave data defition
//===========================================
typedef struct  
{
    INT16 sample_left;         // Signed 16-bit sample
    INT16 sample_right;        // Signed 16-bit sample
} SAMPLE_16_STEREO;

typedef struct  
{
    UINT8 sample_left;         // Unsigned 8-bit sample
    UINT8 sample_right;        // Unsigned 8-bit sample
} SAMPLE_8_STEREO;

typedef struct  
{
    INT16 sample;              // Signed 16-bit sample
} SAMPLE_16_MONO;

typedef struct  
{
    UINT8 sample;              // Unsigned 8-bit sample
} SAMPLE_8_MONO;

typedef union 
{
     SAMPLE_8_MONO m8;
     SAMPLE_16_MONO m16;
     SAMPLE_8_STEREO s8;
     SAMPLE_16_STEREO s16;
}
PCM_SAMPLE, *PPCM_SAMPLE;



//===========================================
// macros
//===========================================
#define BITS_8_TO_16(x)     (INT16)(( (INT32) ((UINT8) x - 128) ) << 8 )
#define BITS_16_TO_8(x)     ((UINT8)((((INT16)(x)) >> 8 ) + 128)) 
#define INC_MODULO(x,y)     (((x)+1) < (y) ? ((x)+1) : 0)


//===========================================
// wave frequency definition
//===========================================
#define  MHZ11_980                  11980000
#define  KHZ08_000                  8000
#define  KHZ11_025                  11025
#define  KHZ12_000                  12000
#define  KHZ16_000                  16000
#define  KHZ22_050                  22050
#define  KHZ24_000                  24000
#define  KHZ32_000                  32000
#define  KHZ44_100                  44100
#define  KHZ48_000                  48000



//===========================================
// buffer definition
//===========================================
#define MAX_BUFFER_PAGE         2

enum
{
	OUTPUT_DMA_PAGE0 = 0,
	OUTPUT_DMA_PAGE1,
	OUTPUT_DMA_PAGE_MAX
};

enum
{
	INPUT_DMA_PAGE0 = 0,
	INPUT_DMA_PAGE1,
	INPUT_DMA_PAGE_MAX
};

#define NB_XMIT_PAGE            MAX_BUFFER_PAGE
#define NB_RECV_PAGE            MAX_BUFFER_PAGE
#define NB_XMIT_BUFF_FIFO       NB_XMIT_PAGE
#define NB_RECV_BUFF_FIFO       NB_RECV_PAGE
#define NUM_DMA_AUDIO_BUFFERS   (NB_XMIT_BUFF_FIFO + NB_RECV_BUFF_FIFO)




#if 0
//===========================================
// function prototypes
//===========================================
EXTERN BOOL         WavHal_AudioMessage(UINT32 Msg, UINT32 Param1, UINT32 Param2);
EXTERN AUDIO_STATE  WavHal_GetInterruptType( VOID );
EXTERN BOOL         WavHal_Init( UINT32 Minor );
EXTERN BOOL         WavHal_WaveOpen( WAPI_INOUT apidir, PWAVEFORMAT pFormat );
EXTERN UINT32       WavHal_WaveOutClose( VOID );
EXTERN UINT32       WavHal_WaveOutContinue( PWAVEHDR pWaveHdr );
EXTERN VOID         WavHal_WaveOutEndOfData( VOID );
EXTERN BOOL         WavHal_WaveOutPause ( VOID );
EXTERN BOOL         WavHal_WaveOutPause ( VOID );
EXTERN BOOL         WavHal_WaveOutRestart( PWAVEHDR pWaveHdr );
EXTERN VOID         WavHal_WaveOutStandby( VOID );
EXTERN VOID         WavHal_WaveOutStart( PWAVEHDR pWaveHdr );
EXTERN VOID         WavHal_WaveOutStop( VOID );

EXTERN VOID         WavHal_WaveInStart( PWAVEHDR pWaveHdr );
EXTERN VOID         WavHal_WaveInContinue( PWAVEHDR pWaveHdr );
EXTERN VOID         WavHal_WaveInStandby( VOID );
EXTERN VOID         WavHal_WaveInStop( PWAVEHDR pWaveHdr );
#endif

#endif  //__WAVE_DRV_H__
